  /*
      f-cpu/ygasm/ygasm.y : syntactical analyser for the F-CPU assembler.
      last hacked Mon Jul 16 02:14:39 by Yann Guidon
      adapted from the PPCM2 which is adapted from GRAZER which is...
  */

  /* the "dangling ELSE" shift-reduce... */
  /*  %expect 1 */
  /* maybe %noassoc would solve the problem ? */

%{

  /* we're limited to long long ints anyway. */
#define YYSTYPE u64

%}


%token YSYMB YNEWSYMB YREG YDB YDD YDQ YDP YDS YORG YALIGN
%token YNUM YAT YSHR YSHL
%token YROP2 YROP2I YLOADCONS YLOADCONSX YMOVE YCMOVE
%token YNOP

 /* fixes the precedence of the operators for the
   evaluation of expr (see at the end of the file) */
%right '='
%left '~'
%left '&'
%left '|'
%left '^'
%left YSHL YSHR
%left '+'
%left '-'
%left '*'
%left '/'
%left '%'
%left STRONG


%%
  /* entry point : a programme is a succession of instructions and pseudo_instructions */
programme : /* nothing */
          | programme pseudo_instruction
          | programme instruction ';' { emit_instruction() }
          ;

pseudo_instruction :
     YNEWSYMB ':' { referenced_symbol->value = current_PC; } /* assign the PC value to the new symbol */
   | YSYMB    '=' expr ';' { referenced_symbol->value = $3 } /* simple assignation */
   | YNEWSYMB '=' expr ';' { referenced_symbol->value = $3 } /* simple assignation (with symbol creation) */
   | YALIGN expr ';'       { align_pc($2); }
   | YORG   expr ';'       { current_PC = $2; }
/*   | data datalist ';' */
;

instruction : instimm161w
            | other_instructions
            ;

 /*       | inst2r1w
          | inst1r1w
          | inst3r1w
          | instimm81r1w
	  ;
 */

instimm161w : opimm16 arg_imm16 ',' YREG { reg_dest = $4 } ;

opimm16 : YLOADCONS  expr ',' {
  opcode = OP_LOADCONS;
  if (($2 >> 4)!=0)
    ygasm_error("invalid shift count for LOADCONS",NULL);
  current_instruction |= ($2 & 15) << 6;
}
        | YLOADCONSX expr ',' {
  opcode = OP_LOADCONSX;
  if (($2 >> 4)!=0)
    ygasm_error("invalid shift count for LOADCONSX",NULL);
  current_instruction |= ($2 & 15) << 6;
}
        ;

arg_imm16 : expr {
  /* verify if we will lose bits in the assignation : */
  s64 a = $1;
  if ((a >> 16)!=(a >> 63)) {
    ygasm_warning("value out of 16-bit range [=>truncating]",NULL);                   
  }
  imm16 = $1;
} ;

other_instructions : YNOP ; /* nothing to do */

   /* All the expressions
      (this is used to compute stuff with labels or constants.)
   */
expr	: '(' expr ')'	          { $$ = $2; }
        | YSYMB	                  { $$ = $1; }
        | YAT	                  { $$ = $1; }
	| YNUM	                  { $$ = $1; }
	| '-' expr %prec STRONG   { $$ = -$2; }
	| expr '-' expr           { $$ = $1 - $3; }
	| expr '+' expr           { $$ = $1 + $3; }
	| expr '*' expr           { $$ = $1 * $3; }
	| expr '/' expr           { $$ = $1 / $3; }
	| expr '%' expr           { $$ = $1 % $3; }
	| expr '&' expr           { $$ = $1 & $3; }
	| expr '|' expr           { $$ = $1 | $3; }
	| expr '^' expr           { $$ = $1 ^ $3; }
	| expr YSHL expr          { $$ = $1 << $3; }
	| expr YSHR expr          { $$ = $1 >> $3; }
	| '~' expr %prec STRONG   { $$ = ~ $2; }
	| '!' expr %prec STRONG   { $$ = ! $2; }
	;

%%
